import configparser
import json
import time

import requests


class EmonPiAdapter:
    """EmonPi handler class used to fetch data using REST API."""

    def __init__(self, config_file: str, *, schema_file: str = ""):

        # Load configuration file
        cfg = configparser.ConfigParser(interpolation=None)
        cfg.read(config_file)
        self.config_file = config_file
        self.endpoint = cfg["Emon"]["endpoint"]
        self.credentials = cfg["Emon"]["bearer_credentials"]
        self.headers = {"Authorization": f"Bearer {self.credentials}"}

        if schema_file:
            with open(schema_file, "r", encoding="utf8") as f_in:
                self.schema = json.load(f_in)

    def _test(self):
        """Checks if connection with given credentials can be established.
        Returns False if request status returns something other than OK.
        """

        try:
            res = requests.get(f"{self.endpoint}/feed/list.json", data={"id": 1}, headers=self.headers)
        except Exception as e:
            print(e)
            return False

        return res.ok

    def get_feed_list(self) -> dict:
        """Fetches the feed list from server.
        Returns the feed list in json format. If something goes wrong, an empty
        dictionary is being returned.
        """

        url = f"{self.endpoint}/feed/list.json"

        res = requests.get(url, headers=self.headers)

        feed_list = {}

        if res.ok:
            feed_list = res.json()

        return feed_list

    def fetch_data(self) -> dict:
        """Fetches only the specified data from server adding an accurate fetch-timestamp.
        Returns the json-like object containing the desired data in the specified
        schema. If something goes wrong, an empty dictionary is being returned.
        """

        feeds = self.get_feed_list()

        # Filter data according to provided schema
        data = {name: feeds[int(id_)]["value"] for id_, name in self.schema.items()}

        # Add a timestamp
        data["timestamp"] = time.time()

        return data



"""A set of predefined observable events"""

import time

from . import Observable


class Timer(Observable):
    """A Timer object notifies all interested clients periodically, in the specified intervals."""

    def __init__(self, interval):
        """Args:
            - interval: int :: Defines the period of busy-waiting in seconds. After `interval` seconds
                               the observers get notified.
        """
        super().__init__()
        self.interval = interval

    def run(self):
        """Repeatedly, busy-waits for the desired amount of time and then notifies the observers"""

        while True:
            time.sleep(self.interval)
            self.notify(time=time.time())


# from datetime import datetime
# from . import Observer
# class DateChange(Observable, Observer):
#     """A DateChange object notifies all interested clients periodically when the date has changed. In order to achieve
#     this, DateChange observes a Timer that defines when the checks will take place."""
#
#     def __init__(self, timer, initial_date=None):
#         self.timer = timer
#         self.current_date = initial_date
#         Observable.__init__(self)
#         Observer.__init__(self, timer)
#
#     def on_notify(self):
#         now = datetime.now()
#         if now != self.current_date:
#             self.current_date = now
#             self.notify(date=now)



"""A simple integration set of utilities connecting EmonPi's ecosystem with CouchDB"""

import os
from .EmonPiAdapter import EmonPiAdapter

_DATA_DIR_NAME = "data"
_CONFIG_FILENAME = "emon-utils.cfg"
_SCHEMA_FILENAME = "schema.json"

PACKAGE_DIR = os.path.dirname(__file__)
DATA_DIR = os.path.join(PACKAGE_DIR, _DATA_DIR_NAME)

# Ensure that the `$PACKAGE/data` dir exists
if not os.path.isdir(DATA_DIR):
    os.mkdir(DATA_DIR)

# Check if config-file lies in directory of execution
# If not, consider working with the global one
if os.path.exists(_CONFIG_FILENAME):
    CONFIG_FILE = os.path.abspath(_CONFIG_FILENAME)
else:
    CONFIG_FILE = os.path.join(DATA_DIR, _CONFIG_FILENAME)

    # If there is no global file, generate a new one
    if not os.path.isfile(CONFIG_FILE):
        from .setup import generate_config
        generate_config(CONFIG_FILE)

# Check if schema-file lies in directory of execution
# If not, consider working with the global one
if os.path.exists(_SCHEMA_FILENAME):
    SCHEMA_FILE = os.path.abspath(_SCHEMA_FILENAME)
else:
    SCHEMA_FILE = os.path.join(DATA_DIR, _SCHEMA_FILENAME)

    # If there is no global file, generate a new one
    if not os.path.isfile(SCHEMA_FILE):
        from .setup import generate_schema
        generate_schema(CONFIG_FILE, SCHEMA_FILE)



"""A simplistic implementation of the Observer Design Pattern."""

from abc import ABC
from abc import abstractmethod


class Observable:
    """Base class for any event generator that would like to massively notify interested observers"""

    def __init__(self):
        self.observers = []

    def register(self, observer):
        """Register an observer to the current observable object

        If the same observer is registered more than once, registration will fail silently.
        """

        if observer not in self.observers:
            self.observers.append(observer)

    def notify(self, *args, **kwargs):
        """Notify all interested (registered) observers"""

        for observer in self.observers:
            observer.on_notify(*args, **kwargs)


class Observer(ABC):
    """Abstract Base Class for any client who should "do something" when an event occurs"""

    def __init__(self, observable):
        observable.register(self)

    @abstractmethod
    def on_notify(self, *args, **kwargs):
        """Default actions that should take place if the desired event occurs"""

        pass



import json

from ..EmonPiAdapter import EmonPiAdapter


def generate_schema(config_file, schema_file):
    emon = EmonPiAdapter(config_file)
    list_ = emon.get_feed_list()

    print("Listing available feeds:\n")

    map_ = {}

    for position_id, feed in enumerate(list_):
        name = feed["name"]
        print(f"#{feed['id']} {name}: {feed['value']}")
        ans = input(f"Rename {name} to: ")
        if ans:
            name = ans
        map_[position_id] = name
        print("---")

    with open(schema_file, "w", encoding="utf8") as f_out:
        json.dump(map_, f_out)

    print(f"File was generated successfully at: {schema_file}!")



import configparser
import getpass


def generate_config(config_file):
    config = configparser.ConfigParser(interpolation=None)
    config["Emon"] = {}
    config["DB"] = {}

    # EmonIO Section
    print("--- Configure EmonIO ---")

    # IP
    default = "127.0.0.1"
    temp = input(f"IP of EmonPi ({default}): ")
    emon_ip = temp if temp else default

    # Bearer Credentials
    bearer_credentials = input("EmonPi Bearer Credentials (only the hash): ")
    if not bearer_credentials:
        bearer_credentials = "REPLACE_ME"
        print("Bearer Credentials cannot be omitted. This will cause trouble!")
        print(f"Please refer to http://{emon_ip}/feed/api, look up for the API-Key")
        print("and replace it in the generated config file.")
        input("Press enter to continue...")

    # DBIO
    print("--- Configure DBIO ---")

    # IP
    default = "127.0.0.1"
    temp = input(f"IP of CouchDB Server ({default}): ")
    db_ip = temp if temp else default

    # Port
    default = "5984"
    temp = input(f"Port of CouchDB Server ({default}): ")
    db_port = temp if temp else default

    # Database Name
    db_name = input("CouchDB Database Name: ")
    if not db_name:
        db_name = "REPLACE_ME"
        print("Database Name cannot be omitted. This will cause trouble!")
        print("Please specify the predefined database")
        print("and replace it in the generated config file.")
        input("Press enter to continue...")

    # Document Name
    document_name = input("CouchDB Document Name: ")
    if not document_name:
        document_name = "REPLACE_ME"
        print("Document Name cannot be omitted. This will cause trouble!")
        print("Please specify the predefined document")
        print("and replace it in the generated config file.")
        input("Press enter to continue...")

    # Username and Password
    username = input("CouchDB Username: ")
    password = getpass.getpass("CouchDB Password: ")

    config["Emon"]["endpoint"] = f"http://{emon_ip}"
    config["Emon"]["bearer_credentials"] = bearer_credentials
    config["DB"]["endpoint"] = f"http://{db_ip}:{db_port}"
    config["DB"]["db_name"] = db_name
    config["DB"]["document_name"] = document_name
    config["DB"]["username"] = username
    config["DB"]["password"] = password

    with open(config_file, "w", encoding="utf8") as f_out:
        config.write(f_out)

    print(f"Config file was successfully generated at: {config_file}")
